import logging
from collections import namedtuple
from unittest.mock import Mock, patch

import pytest

from yapw.decorators import decorate, discard, requeue
from yapw.util import default_decode

# https://pika.readthedocs.io/en/stable/modules/spec.html#pika.spec.Basic.Deliver
Deliver = namedtuple("Deliver", "delivery_tag redelivered routing_key")
# https://pika.readthedocs.io/en/stable/modules/spec.html#pika.spec.BasicProperties
BasicProperties = namedtuple("BasicProperties", "content_type")

logger = logging.getLogger(__name__)


def raises(*args):
    raise RuntimeError("message")


def passes(*args):
    pass


def closes(*args):
    global opened  # noqa: PLW0603
    opened = True
    try:
        raise RuntimeError("message")
    finally:
        opened = False


def finalback(decode, callback, state, channel, method, properties, body):
    def errback(exception):
        logger.warning("errback")

    def finalback():
        logger.warning("finalback")

    decorate(decode, callback, state, channel, method, properties, body, errback, finalback)


@patch("yapw.decorators.nack")
def test_decode_json(nack, caplog):
    caplog.set_level(logging.DEBUG)

    method = Deliver(1, False, "key")
    properties = BasicProperties("application/json")
    callback = Mock()

    discard(default_decode, callback, "state", "channel", method, properties, b'{"message": "value"}')

    callback.assert_called_once_with("state", "channel", method, properties, {"message": "value"})
    nack.assert_not_called()

    assert len(caplog.records) == 1
    assert caplog.records[-1].levelname == "DEBUG"
    assert (
        caplog.records[-1].message
        == 'Received message b\'{"message": "value"}\' with routing key key and delivery tag 1'
    )


@patch("yapw.decorators.nack")
def test_decode_bytes(nack, caplog):
    caplog.set_level(logging.DEBUG)

    method = Deliver(1, False, "key")
    properties = BasicProperties("application/octet-stream")
    callback = Mock()

    discard(default_decode, callback, "state", "channel", method, properties, b"message value")

    callback.assert_called_once_with("state", "channel", method, properties, b"message value")
    nack.assert_not_called()

    assert len(caplog.records) == 1
    assert caplog.records[-1].levelname == "DEBUG"
    assert caplog.records[-1].message == "Received message b'message value' with routing key key and delivery tag 1"


def test_decode_invalid(caplog):
    caplog.set_level(logging.DEBUG)

    state = Mock()
    method = Deliver(1, False, "key")
    properties = BasicProperties("application/json")

    discard(default_decode, passes, state, "channel", method, properties, b"invalid")

    state.connection.ioloop.call_soon_threadsafe.assert_called_once_with(state.interrupt)

    assert len(caplog.records) == 2
    assert caplog.records[0].levelname == "DEBUG"
    assert caplog.records[0].message == "Received message b'invalid' with routing key key and delivery tag 1"
    assert caplog.records[-1].levelname == "ERROR"
    assert caplog.records[-1].message == "b'invalid' can't be decoded, shutting down gracefully"
    assert caplog.records[-1].exc_info


@patch("yapw.decorators.nack")
def test_discard(nack, caplog):
    caplog.set_level(logging.DEBUG)

    method = Deliver(1, False, "key")
    properties = BasicProperties("application/json")

    discard(default_decode, raises, "state", "channel", method, properties, b'"body"')

    nack.assert_called_once_with("state", "channel", 1, requeue=False)

    assert len(caplog.records) == 2
    assert caplog.records[0].levelname == "DEBUG"
    assert caplog.records[0].message == "Received message b'\"body\"' with routing key key and delivery tag 1"
    assert caplog.records[-1].levelname == "ERROR"
    assert caplog.records[-1].message == "Unhandled exception when consuming b'\"body\"', discarding message"
    assert caplog.records[-1].exc_info


@pytest.mark.parametrize(("redelivered", "requeue_kwarg"), [(False, True), (True, False)])
@patch("yapw.decorators.nack")
def test_requeue(nack, redelivered, requeue_kwarg, caplog):
    caplog.set_level(logging.DEBUG)

    method = Deliver(1, redelivered, "key")
    properties = BasicProperties("application/json")

    requeue(default_decode, raises, "state", "channel", method, properties, b'"body"')

    nack.assert_called_once_with("state", "channel", 1, requeue=requeue_kwarg)

    assert len(caplog.records) == 2
    assert caplog.records[0].levelname == "DEBUG"
    assert caplog.records[0].message == "Received message b'\"body\"' with routing key key and delivery tag 1"
    assert caplog.records[-1].levelname == "ERROR"
    assert caplog.records[-1].message == f"Unhandled exception when consuming b'\"body\"' (requeue={requeue_kwarg})"
    assert caplog.records[-1].exc_info


@patch("yapw.decorators.nack")
def test_finally(nack):
    method = Deliver(1, False, "key")
    properties = BasicProperties("application/json")

    discard(default_decode, closes, "state", "channel", method, properties, b'"body"')

    assert opened is False


@patch("yapw.decorators.nack")
def test_finalback_raises(nack, caplog):
    method = Deliver(1, False, "key")
    properties = BasicProperties("application/json")

    finalback(default_decode, raises, "state", "channel", method, properties, b'"body"')

    assert len(caplog.records) == 2
    assert caplog.records[0].levelname == "WARNING"
    assert caplog.records[0].message == "errback"
    assert caplog.records[-1].levelname == "WARNING"
    assert caplog.records[-1].message == "finalback"


@patch("yapw.decorators.nack")
def test_finalback_passes(nack, caplog):
    method = Deliver(1, False, "key")
    properties = BasicProperties("application/json")

    finalback(default_decode, passes, "state", "channel", method, properties, b'"body"')

    assert len(caplog.records) == 1
    assert caplog.records[-1].levelname == "WARNING"
    assert caplog.records[-1].message == "finalback"



import logging
import os
import re
from unittest.mock import patch
from urllib.parse import urlsplit

import pika
import pytest

from yapw.clients import Async

logger = logging.getLogger(__name__)

RABBIT_URL = os.getenv("TEST_RABBIT_URL", "amqp://127.0.0.1")


@patch("yapw.clients.AsyncioConnection")
def test_init_default(connection):
    Async().connect()

    connection.assert_called_once()

    assert connection.call_args[0][0].virtual_host == "/"
    assert connection.call_args[0][0].blocked_connection_timeout == 1800


@patch("yapw.clients.AsyncioConnection")
def test_init_kwargs(connection):
    Async(
        url="https://host:1234/%2Fv?blocked_connection_timeout=10",
        blocked_connection_timeout=300,
    ).connect()

    connection.assert_called_once()

    assert connection.call_args[0][0].virtual_host == "/v"
    assert connection.call_args[0][0].blocked_connection_timeout == 300


@pytest.mark.parametrize(
    ("url", "expected"),
    [
        (
            "amqp://invalid",
            r"AMQPConnectionWorkflowFailed: 1 exceptions in all; last exception - gaierror\(",
        ),
        (
            "amqp://127.0.0.1:1024",
            r"AMQPConnectionError: \(AMQPConnectionWorkflowFailed: 1 exceptions in all; last exception - "
            r"AMQPConnectorSocketConnectError: ConnectionRefusedError\(",
        ),
    ],
)
def test_connection_open_error_bad_host_or_port(url, expected, short_reconnect_delay, caplog):
    caplog.set_level(logging.CRITICAL, logger="pika")
    caplog.set_level(logging.INFO, logger="asyncio")
    caplog.set_level(logging.DEBUG)

    client = Async(durable=False, url=url)
    # Prevent an infinite loop.
    client.stopping = True
    client.start()

    # Channel never opened.
    assert client.connection.is_closed

    assert len(caplog.records) == 1
    for r in caplog.records:
        assert r.levelname == "ERROR"
        assert re.search(r"^Connection failed, retrying in 1s: " + expected, r.message)


@pytest.mark.parametrize("auth", ["invalid:invalid", "guest:invalid"])
def test_connection_open_error_bad_username_or_password(auth, short_reconnect_delay, caplog):
    caplog.set_level(logging.CRITICAL, logger="pika")
    caplog.set_level(logging.INFO, logger="asyncio")
    caplog.set_level(logging.DEBUG)

    parsed = urlsplit(RABBIT_URL)
    parsed = parsed._replace(netloc=f"{auth}@{parsed.hostname}:{parsed.port or 5672}")

    client = Async(durable=False, url=parsed.geturl())
    # Prevent an infinite loop.
    client.stopping = True
    client.start()

    # Channel never opened.
    assert client.connection.is_closed

    assert len(caplog.records) == 1
    assert [(r.levelname, r.message) for r in caplog.records] == [
        (
            "ERROR",
            "Stopping: ProbableAuthenticationError: Client was disconnected at a connection stage indicating a "
            "probable authentication error: (\"ConnectionClosedByBroker: (403) 'ACCESS_REFUSED - Login was refused "
            "using authentication mechanism PLAIN. For details see the broker logfile.'\",)",
        )
    ]


def test_connection_open_error_bad_virtual_host(short_reconnect_delay, caplog):
    caplog.set_level(logging.CRITICAL, logger="pika")
    caplog.set_level(logging.INFO, logger="asyncio")
    caplog.set_level(logging.DEBUG)

    parsed = urlsplit(RABBIT_URL)
    parsed = parsed._replace(path="/%2Finvalid")

    client = Async(durable=False, url=parsed.geturl())
    # Prevent an infinite loop.
    client.stopping = True
    client.start()

    # Channel never opened.
    assert client.connection.is_closed

    assert len(caplog.records) == 1
    assert [(r.levelname, r.message) for r in caplog.records] == [
        (
            "ERROR",
            "Stopping: ProbableAccessDeniedError: Client was disconnected at a connection stage indicating a probable "
            "denial of access to the specified virtual host: (\"ConnectionClosedByBroker: (530) 'NOT_ALLOWED - vhost "
            "/invalid not found'\",)",
        )
    ]


def test_connection_close(short_reconnect_delay, caplog):
    class Client(Async):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.overridden = False

        def exchange_ready(self):
            self.interrupt()
            if not self.overridden:  # use the second branch of connection_close_callback() once
                self.stopping = False
                self.overridden = True

    client = Client(durable=False, url=RABBIT_URL)
    client.start()

    assert client.channel.is_closed
    assert client.connection.is_closed

    assert len(caplog.records) == 3
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("WARNING", "Channel 1 was closed: ChannelClosedByClient: (200) 'Normal shutdown'"),
        ("WARNING", "Connection closed, reconnecting in 1s: ConnectionClosedByClient: (200) 'Normal shutdown'"),
        ("WARNING", "Channel 1 was closed: ChannelClosedByClient: (200) 'Normal shutdown'"),
    ]


def test_exchangeok_default(short_timer, caplog):
    caplog.set_level(logging.DEBUG)

    class Client(Async):
        def exchange_ready(self):
            logger.info("stop")

    client = Client(durable=False, url=RABBIT_URL)
    client.start()

    assert client.channel.is_closed
    assert client.connection.is_closed

    assert len(caplog.records) == 3
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("INFO", "stop"),
        ("INFO", "Received SIGINT, shutting down gracefully"),
        ("WARNING", "Channel 1 was closed: ChannelClosedByClient: (200) 'Normal shutdown'"),
    ]


@pytest.mark.parametrize("exchange_type", [pika.exchange_type.ExchangeType.direct, "direct"])
def test_exchangeok_kwargs(exchange_type, short_timer, caplog):
    caplog.set_level(logging.DEBUG)

    class Client(Async):
        def exchange_ready(self):
            logger.info("stop")

    client = Client(durable=False, url=RABBIT_URL, exchange="yapw_test", exchange_type=exchange_type)
    client.start()

    assert client.ready is True
    assert client.channel.is_closed
    assert client.connection.is_closed

    assert len(caplog.records) == 3
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("INFO", "stop"),
        ("INFO", "Received SIGINT, shutting down gracefully"),
        ("WARNING", "Channel 1 was closed: ChannelClosedByClient: (200) 'Normal shutdown'"),
    ]



import functools
import json
import logging
import os
import signal
import time
from contextlib import contextmanager
from threading import Timer

from yapw.clients import Blocking
from yapw.methods import ack, nack, publish

DELAY = 0.05
RABBIT_URL = os.getenv("TEST_RABBIT_URL", "amqp://127.0.0.1")

logger = logging.getLogger(__name__)


def kill(signum):
    os.kill(os.getpid(), signum)
    # The signal should be handled once.
    os.kill(os.getpid(), signum)


@contextmanager
def timed(interval):
    timer = Timer(interval, functools.partial(kill, signal.SIGINT))
    timer.start()
    try:
        yield
    finally:
        timer.cancel()


def blocking(**kwargs):
    return Blocking(durable=False, url=RABBIT_URL, exchange="yapw_test", **kwargs)


def encode(message):
    if not isinstance(message, bytes):
        return json.dumps(message, separators=(",", ":")).encode()
    return message


def decode(index, body, content_type):
    return body.decode()[index]


# Consumer callbacks.
def sleeper(state, channel, method, properties, body):
    logger.info("Sleep")
    time.sleep(DELAY * 2)
    logger.info("Wake!")
    ack(state, channel, method.delivery_tag)


def raiser(state, channel, method, properties, body):
    raise RuntimeError("message")


def ack_warner(state, channel, method, properties, body):
    logger.warning(body)
    ack(state, channel, method.delivery_tag)


def nack_warner(state, channel, method, properties, body):
    logger.warning(body)
    nack(state, channel, method.delivery_tag)


def writer(state, channel, method, properties, body):
    publish(state, channel, {"message": "value"}, "n")
    ack(state, channel, method.delivery_tag)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent.parent))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Yet Another Pika Wrapper"
copyright = "2021, Open Contracting Partnership"
author = "Open Contracting Partnership"

version = "0.2.1"
release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx_design",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = []

# -- Extension configuration -------------------------------------------------

autodoc_default_options = {
    "members": None,
    "member-order": "bysource",
}
autodoc_typehints = "description"
autodoc_type_aliases = {
    "ConsumerCallback": "yapw.types.ConsumerCallback",
    "Decode": "yapw.types.Decode",
    "Decorator": "yapw.types.Decorator",
    "Encode": "yapw.types.Encode",
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pika": ("https://pika.readthedocs.io/en/stable/", None),
}



# Copied and adapted from https://github.com/scrapy/scrapy/blob/master/scrapy/utils/ossignal.py
import signal

signal_names = {}
for signame in dir(signal):
    if signame.startswith("SIG") and not signame.startswith("SIG_"):
        signum = getattr(signal, signame)
        if isinstance(signum, int):
            signal_names[signum] = signame



import pika
import pytest

from yapw.clients import Base
from yapw.util import default_decode, default_encode


def dumps(message, content_type):
    return b"overridden"


def test_init_default():
    client = Base()

    expected = pika.URLParameters("amqp://127.0.0.1")
    expected.blocked_connection_timeout = 1800

    assert client.parameters == expected
    assert client.durable is True
    assert client.exchange == ""
    assert client.exchange_type == "direct"
    assert client.prefetch_count == 1
    assert client.decode == default_decode
    assert client.encode == default_encode
    assert client.content_type == "application/json"
    assert client.delivery_mode == 2
    assert client.routing_key_template == "{exchange}_{routing_key}"

    assert client.format_routing_key("test") == "_test"


def test_init_kwargs():
    client = Base(
        url="amqp://localhost",
        blocked_connection_timeout=0,
        durable=False,
        exchange="exch",
        exchange_type="fanout",
        prefetch_count=10,
        decode=default_decode,
        encode=dumps,
        content_type="application/octet-stream",
        routing_key_template="{routing_key}_{exchange}",
    )

    expected = pika.URLParameters("amqp://localhost")
    expected.blocked_connection_timeout = 0

    assert client.parameters == expected
    assert client.durable is False
    assert client.exchange == "exch"
    assert client.exchange_type == "fanout"
    assert client.prefetch_count == 10
    assert client.decode == default_decode
    assert client.encode == dumps
    assert client.content_type == "application/octet-stream"
    assert client.delivery_mode == 1
    assert client.routing_key_template == "{routing_key}_{exchange}"

    assert client.format_routing_key("test") == "test_exch"


def test_format_routing_key_invalid():
    client = Base(routing_key_template="{invalid}_{routing_key}")

    with pytest.raises(KeyError):
        client.format_routing_key("test")



"""
This script can be used to test error handling.
"""

import logging
import sys

from yapw.clients import AsyncConsumer, Blocking
from yapw.decorators import requeue

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def callback(state, channel, method, properties, body):
    raise RuntimeError("message")


def main():
    kwargs = {"durable": False, "exchange": "yapw_development"}
    consumer_kwargs = {"on_message_callback": callback, "queue": "raise", "decorator": requeue}
    if len(sys.argv) > 1 and sys.argv[1] == "Blocking":
        print("Blocking consumer")
        client = Blocking(**kwargs)
        client.consume(**consumer_kwargs)
    else:
        print("Asynchronous consumer")
        client = AsyncConsumer(**kwargs, **consumer_kwargs)
        client.start()


if __name__ == "__main__":
    main()



"""
This script can be used to test publishing (like the sample code in the documentation).
"""

import logging

from yapw.clients import Async

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class Publisher(Async):
    def exchange_ready(self):
        self.publish({"message": "value"}, routing_key="messages")
        self.interrupt()


def main():
    publisher = Publisher(durable=False, exchange="yapw_development")
    publisher.start()


if __name__ == "__main__":
    main()



import functools
import logging
import signal

import pytest

from tests import DELAY, ack_warner, blocking, decode, encode, kill, nack_warner, raiser, sleeper, writer
from yapw.decorators import discard, requeue

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    ("signum", "signame"),
    [(signal.SIGINT, "SIGINT"), (signal.SIGTERM, "SIGTERM")],
)
def test_shutdown(signum, signame, message, caplog):
    caplog.set_level(logging.INFO)

    consumer = blocking()
    consumer.connection.call_later(DELAY, functools.partial(kill, signum))
    consumer.consume(sleeper, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 3
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("INFO", "Sleep"),
        ("INFO", f"Received {signame}, shutting down gracefully"),
        ("INFO", "Wake!"),
    ]


def test_decode_valid(short_message, caplog):
    consumer = blocking(decode=functools.partial(decode, 0))
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(ack_warner, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 1
    assert [(r.levelname, r.message) for r in caplog.records] == [("WARNING", "1")]


def test_decode_invalid(short_message, caplog):
    caplog.set_level(logging.INFO)

    consumer = blocking(decode=functools.partial(decode, 10))  # IndexError
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(ack_warner, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 1
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"{encode(short_message)} can't be decoded, shutting down gracefully", False),
    ]


def test_decode_raiser(message, caplog):
    caplog.set_level(logging.INFO)

    consumer = blocking(decode=raiser)
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(ack_warner, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 1
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"{encode(message)} can't be decoded, shutting down gracefully", False),
    ]


def test_halt(message, caplog):
    caplog.set_level(logging.INFO)

    consumer = blocking()
    consumer.connection.call_later(30, functools.partial(kill, signal.SIGINT))  # in case not halted
    consumer.consume(raiser, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 1
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"Unhandled exception when consuming {encode(message)}, shutting down gracefully", False),
    ]


def test_discard(message, caplog):
    caplog.set_level(logging.INFO)

    consumer = blocking()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(raiser, "q", decorator=discard)

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 2
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"Unhandled exception when consuming {encode(message)}, discarding message", False),
        ("INFO", "Received SIGINT, shutting down gracefully", True),
    ]


def test_requeue(message, caplog):
    caplog.set_level(logging.INFO)

    consumer = blocking()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(raiser, "q", decorator=requeue)

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 3
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"Unhandled exception when consuming {encode(message)} (requeue=True)", False),
        ("ERROR", f"Unhandled exception when consuming {encode(message)} (requeue=False)", False),
        ("INFO", "Received SIGINT, shutting down gracefully", True),
    ]


def test_publish(message, caplog):
    caplog.set_level(logging.DEBUG)

    consumer = blocking()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(writer, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 5
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("DEBUG", "Consuming messages on channel 1 from queue yapw_test_q"),
        ("DEBUG", f"Received message {encode(message)} with routing key yapw_test_q and delivery tag 1"),
        (
            "DEBUG",
            "Published message {'message': 'value'} on channel 1 to exchange yapw_test with routing key yapw_test_n",
        ),
        ("DEBUG", "Ack'd message on channel 1 with delivery tag 1"),
        ("INFO", "Received SIGINT, shutting down gracefully"),
    ]


def test_consume_declares_queue(caplog):
    declarer = blocking()
    declarer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    declarer.consume(raiser, "q")

    publisher = blocking()
    publisher.publish({"message": "value"}, "q")

    consumer = blocking()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(nack_warner, "q")

    publisher.channel.queue_purge("yapw_test_q")
    publisher.close()

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) > 1
    assert all(r.levelname == "WARNING" and r.message == "{'message': 'value'}" for r in caplog.records)


def test_consume_declares_queue_routing_keys(caplog):
    declarer = blocking()
    declarer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    declarer.consume(raiser, "q", ["r", "k"])

    publisher = blocking()
    publisher.publish({"message": "r"}, "r")
    publisher.publish({"message": "k"}, "k")

    consumer = blocking()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(ack_warner, "q", ["r", "k"])

    publisher.channel.queue_purge("yapw_test_q")
    publisher.close()

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 2
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("WARNING", "{'message': 'r'}"),
        ("WARNING", "{'message': 'k'}"),
    ]



"""
Decorators to be used with consumer callbacks.

A message must be ack'd or nack'd if using `consumer prefetch <https://www.rabbitmq.com/consumer-prefetch.html>`__,
because otherwise `RabbitMQ stops delivering messages <https://www.rabbitmq.com/confirms.html#channel-qos-prefetch>`__.
The decorators help to ensure that, in case of error, either the message is nack'd or the process is halted.

:func:`~yapw.decorators.halt` is the default decorator. It stops the consumer and halts the process, so that an
administrator can decide when it is appropriate to restart it.

The other decorators require more care. For example, if a callback inserts messages into a database, and the database
is down, but this exception isn't handled by the callback, then the :func:`~yapw.decorators.discard` or
:func:`~yapw.decorators.requeue` decorators would end up nack'ing all messages in the queue.

Decorators look like this (see the :func:`~yapw.decorators.decorate` function for context):

.. code-block:: python

   from yapw.decorators import decorate


   def myfunction(decode, callback, state, channel, method, properties, body):
       def errback(exception):
           # do something, like halting the process or nack'ing the message

       decorate(decode, callback, state, channel, method, properties, body, errback)
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from yapw.methods import add_callback_threadsafe, nack

if TYPE_CHECKING:
    from collections.abc import Callable

    import pika

    from yapw.types import ConsumerCallback, Decode, State

logger = logging.getLogger(__name__)


def decorate(
    decode: Decode,
    callback: ConsumerCallback,
    state: State[Any],
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
    errback: Callable[[Exception], None],
    finalback: Callable[[], None] | None = None,
) -> None:
    """
    Use this function to define your own decorators.

    Decode the message ``body`` using the ``decode`` function, and call the consumer ``callback``.

    If the ``callback`` function raises an exception, call the ``errback`` function. In any case, call the
    ``finalback`` function after calling the ``callback`` function.

    If the ``decode`` function raises an exception, shut down the client in the main thread.

    .. seealso::

       :mod:`yapw.clients` for details on the consumer callback function signature.
    """
    logger.debug(
        "Received message %s with routing key %s and delivery tag %s", body, method.routing_key, method.delivery_tag
    )
    try:
        message = decode(body, properties.content_type)
        try:
            callback(state, channel, method, properties, message)
        except Exception as exception:  # noqa: BLE001
            errback(exception)
        finally:
            if finalback:
                finalback()
    except Exception:
        logger.exception("%r can't be decoded, shutting down gracefully", body)
        add_callback_threadsafe(state.connection, state.interrupt)


# https://stackoverflow.com/a/7099229/244258
def halt(
    decode: Decode,
    callback: ConsumerCallback,
    state: State[Any],
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
) -> None:
    """
    If the ``callback`` function raises an exception, shut down the client in the main thread, without acknowledgment.
    """

    def errback(_exception: Exception) -> None:
        logger.exception("Unhandled exception when consuming %r, shutting down gracefully", body)
        add_callback_threadsafe(state.connection, state.interrupt)

    decorate(decode, callback, state, channel, method, properties, body, errback)


def discard(
    decode: Decode,
    callback: ConsumerCallback,
    state: State[Any],
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
) -> None:
    """If the ``callback`` function raises an exception, nack the message, without requeueing."""

    def errback(_exception: Exception) -> None:
        logger.exception("Unhandled exception when consuming %r, discarding message", body)
        nack(state, channel, method.delivery_tag, requeue=False)

    decorate(decode, callback, state, channel, method, properties, body, errback)


def requeue(
    decode: Decode,
    callback: ConsumerCallback,
    state: State[Any],
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
) -> None:
    """If the ``callback`` function raises an exception, nack the message, requeueing it unless it was redelivered."""

    def errback(_exception: Exception) -> None:
        requeue = not method.redelivered
        logger.exception("Unhandled exception when consuming %r (requeue=%r)", body, requeue)
        nack(state, channel, method.delivery_tag, requeue=requeue)

    decorate(decode, callback, state, channel, method, properties, body, errback)
